/*****************************************************************************/
/*
                                   CGI.c

Provide generalized CGI scripting support.

Creates a buffer containing either, the DCL commands to create CGI variables at
the VMS command line (e.g. for standard CGI scripts), or a stream of
'name=value' pairs for the CGI variables (e.g. for CGIplus scripts). 

The buffer is dynamically allocated and contains a series of sequential records
comprising a word (16 bits) with the length of a following (varying)
null-terminated string (including the null character).  The end of the series
is indicated by a zero length record.

This buffer is scanned from from first to last and the strings passed to the
script 'input' stream as appropriate.


CGI VARIABLES
-------------
Most of these CGI variable names are those supported by the INTERNET-DRAFT
authored by D.Robinson (drtr@ast.cam.ac.uk), 8 January 1996, plus some
"convenience" variables, breaking the query string into its components (KEY_,
FORM_, etc.) The CGI symbols (CERN/VMS-HTTPd-like DCL symbols instead of Unix
environment variables) are created by the SYS$COMMAND stream of each
subprocess before the script DCL procedure is invoked. By default each
variable name is prefixed by "WWW_" (similar to CERN HTTPd), although this can
be modified at the command line when starting the server.

     VARIABLE NAME            DESCRIPTION                            "standard"
     -------------            -----------                            ----------

  (if the request is not authorized only AUTH_TYPE will exist)
  o  AUTH_ACCESS ............ "READ" or "READ+WRITE"                 [no]
  o  AUTH_AGENT ............. indicates authentication agent         [no]
  o  AUTH_GROUP ............. path authorization group               [no]
  o  AUTH_PASSWORD .......... only if "EXTERNAL" realm               [no]
  o  AUTH_REALM ............. authentication realm                   [no]
  o  AUTH_REALM_DESCRIPTION . realm description                      [no]
  o  AUTH_TYPE .............. "BASIC" or "DIGEST" (or empty)         [yes]
  o  AUTH_USER .............. details of user                        [no]

  o  CONTENT_LENGTH ......... "Content-Length:" from header          [yes]
  o  CONTENT_TYPE ........... "Content-Type:" from header            [yes]

  o  DOCUMENT_ROOT .......... base directory for serving files       [apache]

  o  FORM_field ............. query "&" separated form elements      [no]

  o  GATEWAY_INTERFACE ...... "CGI/1.1"                              [yes]

  o  HTTP_ACCEPT ............ list of browser-accepted content types [yes]
  O  HTTP_ACCEPT_ENCODING ... list of browser-accepted encodings     [yes]
  o  HTTP_ACCEPT_CHARSET .... list of browser-accepted character set [yes]
  o  HTTP_ACCEPT_LANGUAGE ... list of browser-accepted languages     [yes]
  o  HTTP_AUTHORIZATION ..... only if "EXTERNAL" realm               [yes]
  o  HTTP_COOKIE ............ any cookie sent by the client          [yes]
  o  HTTP_FORWARDED ......... list of proxy/gateway hosts            [yes]
  o  HTTP_HOST .............. destination host name/port             [yes]
  o  HTTP_IF_NOT_MODIFIED ... GMT time string                        [yes]
  o  HTTP_PRAGMA ............ any pragma directive of header         [yes]
  o  HTTP_REFERER ........... source document URL for this request   [yes]
  o  HTTP_USER_AGENT ........ client/browser identification string   [yes]

  o  KEY_n .................. query string "+" separated elements    [no]
  o  KEY_COUNT .............. number of "+" separated elements       [no]

  o  PATH_INFO .............. virtual path of data requested in URL  [yes]
  o  PATH_ODS ............... on-disk structure of path (0, 2 or 5)  [no]
  o  PATH_TRANSLATED ........ VMS file path of data requested in URL [yes]

  o  QUERY_STRING ........... string following "?" in URL            [yes]

  o  REMOTE_ADDR ............ IP host address of HTTP client         [yes]
  o  REMOTE_HOST ............ IP host name of HTTP client            [yes]
  o  REMOTE_PORT ............ IP port of HTTP client                 [no]
  o  REMOTE_USER ............ authenticated username (or empty)      [yes]
  o  REQUEST_CHARSET ........ charset SET by mapping rule            [no]
  o  REQUEST_CONTENT_TYPE ... content-type SET by mapping rule       [no]
  o  REQUEST_METHOD ......... "GET", "PUT", etc.                     [yes]
  o  REQUEST_SCHEME ......... "http:" or "https:"                    [no]
  o  REQUEST_TIME_GMT ....... request GMT time                       [no]
  o  REQUEST_TIME_LOCAL ..... request local time                     [no]
  o  REQUEST_URI ............ un-encoded path[?query-string]         [apache]

  o  SCRIPT_NAME ............ name of script (e.g. "/query")         [yes]
  o  SCRIPT_FILENAME ........ (e.g. "cgi-bin:[000000]test.com")      [apache]
  o  SERVER_ADDR ............ IP host address of server system       [no]
  o  SERVER_ADMIN ........... "webmaster@site.domain"                [apache]
  o  SERVER_CHARSET ......... default charset (e.g. "ISO-8895-1")    [no]
  o  SERVER_GMT ............. offset from GMT time (e.g. "+09:30)    [no]
  o  SERVER_NAME ............ IP host name of server system          [yes]
  o  SERVER_PROTOCOL ........ HTTP protocol version ("HTTP/1.0")     [yes]
  o  SERVER_PORT ............ IP port request was received on        [yes]
  o  SERVER_SOFTWARE ........ software ID of the HTTPD daemon        [yes]
  o  SERVER_SIGNATURE ....... "server host port"                     [apache]

  o  UNIQUE_ID .............. a "unqiue" request ID                  [apache]

  o  some SSL related CGI variable can be generated when SSL is available
     (see SESOLA.C module)

NOTE: If script portability is a concern then confine use to the "standard" CGI
variables.  For WASD-specific scripts the extension variables may be used.  Not
all variables will be present for all requests.


CGI VARIABLES USING DCL SYMBOLS
-------------------------------
Due to the mechanism used to create DCL symbols for the CGI variables (for
standard CGI) the variable value is limited to 1005 characters minus the length
of the DCL symbol name assignment (e.g. "WWW_QUERY_STRING=A+B+C+D"), or in
round terms approximately 980 characters maximum.  This 1005 appears to be a
"magic number" for the DCL command line (VAX VMS V7.2 anyway).


CGI VARIABLES USING CGIPLUS STREAM
----------------------------------
The maximum 'name=value' length for a CGI variable conveyed using the CGIPLUSIN
stream is limited by the size of of the 'DclCgiPlusInSize' global storage
value which is in turn set by the [BufferSizeDclCgiPlusIn] configuration
directive.  Hence if this is set to 3072 bytes (the current default) then the
maximum 'name=value' length is the that value minus 3 bytes.  This value is
also the total mailbox buffer space allocated for subprocess IPC and so is also
the limit on all variables!  So, in the circumstance where the largest variable
needs to be 4000 bytes in length, and the rest total up to another 3000 bytes,
the [BufferSizeDclCgiPlusIn] is suggested to be 8192 bytes.  When this value is
adjusted it is recommended to review the HTTP server account's BYTLM quota.


VERSION HISTORY
---------------
10-MAY-2000  MGD  refined CgiOutput() and CgiHeader()
08-APR-2000  MGD  CGI binary header response can now be processed,
                  (some) VMS Apache compatibility
31-DEC-1999  MGD  PATH_ODS variable (to indicate on-disk file system)
17-DEC-1999  MGD  bugfix; quote double-up in CgiVariable()
27-NOV-1999  MGD  rework DCL variable stream,
                  remove sys$fao() from stream CGI variable creation,
                  avoid "''" substitution creating DCL symbol CGI variables
20-NOV-1999  MGD  end-of-header blank line may now be an empty record,
                  or a record containing a single '\n' or '\r\n' sequence
06-NOV-1999  MGD  allow slightly more latitude with form fields, no "=value"
                  (i.e. name terminated by '&') and back-to-back '&'
14-SEP-1999  MGD  modify CgiOutput() for delayed CGI "Content-Type:"
28-AUG-1999  MGD  support CGIplus "agents", added AUTH_USER
30-MAY-1999  MGD  add SSL variables (via call to SesolaCgiGenerateVariables())
21-FEB-1999  MGD  change authorization CGI variables
28-DEC-1998  MGD  added HTTP_ACCEPT_ENCODING
07-NOV-1998  MGD  WATCH facility
17-OCT-1998  MGD  CgiUrlDecodeString(),
                  SERVER_CHARSET, REQUEST_CHARSET, REQUEST_CONTENT_TYPE
28-APR-1998  MGD  CGI variable memory allocation changed in support of SSI
03-APR-1998  MGD  bugfix; extra quotes generated in form field value
19-MAR-1998  MGD  suppress 'Authorization:' unless "external" authorization
06-DEC-1997  MGD  functionality unbundled from DCL.c, generalized for version 5
*/
/*****************************************************************************/

/* standard C header files */
#include <ctype.h>
#include <stdio.h>

/* VMS related header files */
#include <descrip.h>
#include <dvidef.h>
#include <ssdef.h>
#include <stsdef.h>

/* application header files */
#include "wasd.h"

#define WASD_MODULE "CGI"

/***********/
/* defines */
/***********/

#define CLI$_BUFOVF 229400

/******************/
/* global storage */
/******************/

char  ErrorCgiNotStrict [] = "NOT a strict CGI response!!";

/********************/
/* external storage */
/********************/

#ifdef DBUG
extern boolean Debug;
#else
#define Debug 0 
#endif

extern boolean  OdsExtended;

extern int  DclCgiPlusInSize,
            DclSysCommandSize,
            ServerPort,
            WatchEnabled;

unsigned long HttpdProcessId;

extern char  DclCgiVariablePrefix[],
             HttpProtocol[],
             SoftwareID[],
             TimeGmtString[];

extern struct ConfigStruct Config;
extern struct MsgStruct Msgs;

/*****************************************************************************/
/*
Creates a buffer containing a series of null-terminated strings, terminated by
a null-string (two successive null characters).  Each of the strings contains
either DCL command(s) to create a DCL symbol (CGI variable) using the CLI, or
an equate separated 'name=value' pair used for CGIplus variable streams, etc.
*/ 

CgiGenerateVariables
(
struct RequestStruct *rqptr,
int VarType
)
{
   static $DESCRIPTOR (NumberFaoDsc, "!UL\0");
   static $DESCRIPTOR (StatusValueFaoDsc, "%X!8UL\0");

   register char  c;
   register char  *cptr, *sptr, *zptr;

   int  status,
        Count,
        KeyCount;
   char  FormFieldName [256],
         LocalDateTime [48],
         String [1024];
   $DESCRIPTOR (StringDsc, String);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CgiGenerateVariables()\n");

   if (rqptr->WatchItem &&
       (WatchEnabled & WATCH_CGI))
   {
      if (VarType == CGI_VARIABLE_STREAM)
         WatchThis (rqptr, FI_LI, WATCH_CGI, "VARIABLES stream");
      else
         WatchThis (rqptr, FI_LI, WATCH_CGI, "VARIABLES dcl");
   }

   if (rqptr->rqCgi.BufferLength)
   {
      /* free previous variables and storage */
      if (rqptr->rqCgi.BufferPtr != NULL)
         VmFreeFromHeap (rqptr, rqptr->rqCgi.BufferPtr, FI_LI); 
      rqptr->rqCgi.BufferLength = 0;
      rqptr->rqCgi.BufferPtr = rqptr->rqCgi.BufferCurrentPtr = NULL;
   }

   if (VarType == CGI_VARIABLE_STREAM)
      CgiVariableBufferMemory (rqptr, DclCgiPlusInSize*2);
   else
      CgiVariableBufferMemory (rqptr, DclSysCommandSize);

   HttpLocalTimeString (LocalDateTime, &rqptr->rqTime.Vms64bit);

   if (rqptr->rqAuth.Type[0])
   {
      if (rqptr->rqAuth.SourceRealm)
      { 
         if (VMSnok (status =
             CgiVariable (rqptr, "AUTH_ACCESS", 
                          AuthCanString (rqptr->rqAuth.RequestCan,
                                         AUTH_CAN_FORMAT_LONG),
                          VarType)))
            return (status);
      }

      if (rqptr->rqAuth.AgentParameterPtr != NULL &&
          rqptr->rqAuth.AgentParameterPtr[0])
      { 
         if (VMSnok (status =
             CgiVariable (rqptr, "AUTH_AGENT",
                          rqptr->rqAuth.AgentParameterPtr, VarType)))
            return (status);
      }

      if (rqptr->rqAuth.GroupWritePtr[0])
      {
         if (VMSnok (status =
             CgiVariable (rqptr, "AUTH_GROUP",
                          rqptr->rqAuth.GroupWritePtr, VarType)))
            return (status);
      }
      else
      {
         if (VMSnok (status =
             CgiVariable (rqptr, "AUTH_GROUP",
                          rqptr->rqAuth.GroupReadPtr, VarType)))
            return (status);
      }

      if ((rqptr->rqAuth.AgentParameterPtr != NULL &&
           rqptr->rqAuth.AgentParameterPtr[0]) ||
          rqptr->rqAuth.SourceRealm == AUTH_SOURCE_EXTERNAL)
      {
         if (VMSnok (status =
             CgiVariable (rqptr, "AUTH_PASSWORD",
                          rqptr->RemoteUserPassword, VarType)))
            return (status);
      }

      if (VMSnok (status =
          CgiVariable (rqptr, "AUTH_REALM", rqptr->rqAuth.RealmPtr, VarType)))
         return (status);

      if (VMSnok (status =
          CgiVariable (rqptr, "AUTH_REALM_DESCRIPTION",
                       rqptr->rqAuth.RealmDescrPtr, VarType)))
         return (status);
   }

   if (VMSnok (status =
       CgiVariable (rqptr, "AUTH_TYPE", rqptr->rqAuth.Type, VarType)))
      return (status);

   if (rqptr->rqAuth.Type[0])
   {
      if (VMSnok (status =
          CgiVariable (rqptr, "AUTH_USER", rqptr->rqAuth.UserDetailsPtr, VarType)))
         return (status);
   }

   sys$fao (&NumberFaoDsc, 0, &StringDsc, rqptr->rqHeader.ContentLength);
   if (VMSnok (status =
       CgiVariable (rqptr, "CONTENT_LENGTH", String, VarType)))
      return (status);

   if (rqptr->rqHeader.ContentTypePtr != NULL)
   {
      if (VMSnok (status =
          CgiVariable (rqptr, "CONTENT_TYPE",
                       rqptr->rqHeader.ContentTypePtr, VarType)))
         return (status);
   }
   else
   if (rqptr->rqContentInfo.ContentTypePtr != NULL)
   {
      if (VMSnok (status =
          CgiVariable (rqptr, "CONTENT_TYPE",
                       rqptr->rqContentInfo.ContentTypePtr, VarType)))
         return (status);
   }
   else
   {
      if (VMSnok (status =
          CgiVariable (rqptr, "CONTENT_TYPE", "", VarType)))
         return (status);
   }

   /* Apache-like - just empty for WASD */
   if (VMSnok (status =
       CgiVariable (rqptr, "DOCUMENT_ROOT", "", VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "GATEWAY_INTERFACE", "CGI/1.1", VarType)))
      return (status);

   if (rqptr->rqHeader.AcceptPtr != NULL)
      if (VMSnok (status =
          CgiVariable (rqptr, "HTTP_ACCEPT", rqptr->rqHeader.AcceptPtr, VarType)))
      return (status);

   if (rqptr->rqHeader.AcceptCharsetPtr != NULL)
      if (VMSnok (status =
          CgiVariable (rqptr, "HTTP_ACCEPT_CHARSET",
                       rqptr->rqHeader.AcceptCharsetPtr, VarType)))
      return (status);

   if (rqptr->rqHeader.AcceptEncodingPtr != NULL)
      if (VMSnok (status =
          CgiVariable (rqptr, "HTTP_ACCEPT_ENCODING",
                       rqptr->rqHeader.AcceptEncodingPtr, VarType)))
      return (status);

   if (rqptr->rqHeader.AcceptLangPtr != NULL)
      if (VMSnok (status =
          CgiVariable (rqptr, "HTTP_ACCEPT_LANGUAGE",
                       rqptr->rqHeader.AcceptLangPtr, VarType)))
         return (status);

   if (rqptr->rqAuth.SourceRealm == AUTH_SOURCE_EXTERNAL &&
       rqptr->rqHeader.AuthorizationPtr != NULL)
      if (VMSnok (status =
          CgiVariable (rqptr, "HTTP_AUTHORIZATION",
                       rqptr->rqHeader.AuthorizationPtr, VarType)))
         return (status);

   if (rqptr->rqHeader.CookiePtr != NULL)
      if (VMSnok (status =
          CgiVariable (rqptr, "HTTP_COOKIE", rqptr->rqHeader.CookiePtr, VarType)))
         return (status);

   if (rqptr->rqHeader.ForwardedPtr != NULL)
      if (VMSnok (status =
          CgiVariable (rqptr, "HTTP_FORWARDED",
                       rqptr->rqHeader.ForwardedPtr, VarType)))
         return (status);

   if (rqptr->rqHeader.HostPtr != NULL)
      if (VMSnok (status =
          CgiVariable (rqptr, "HTTP_HOST", rqptr->rqHeader.HostPtr, VarType)))
         return (status);

   if (rqptr->rqHeader.IfModifiedSincePtr != NULL)
      if (VMSnok (status =
          CgiVariable (rqptr, "HTTP_IF_MODIFIED_SINCE",
                       rqptr->rqHeader.IfModifiedSincePtr, VarType)))
         return (status);

   if (rqptr->rqHeader.PragmaPtr != NULL)
      if (VMSnok (status =
          CgiVariable (rqptr, "HTTP_PRAGMA", rqptr->rqHeader.PragmaPtr, VarType)))
         return (status);

   if (rqptr->rqHeader.RefererPtr != NULL)
      if (VMSnok (status =
          CgiVariable (rqptr, "HTTP_REFERER", rqptr->rqHeader.RefererPtr, VarType)))
         return (status);

   if (rqptr->rqHeader.UserAgentPtr != NULL)
      if (VMSnok (status =
          CgiVariable (rqptr, "HTTP_USER_AGENT",
                       rqptr->rqHeader.UserAgentPtr, VarType)))
         return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "PATH_INFO", rqptr->rqHeader.PathInfoPtr, VarType)))
      return (status);

   if (OdsExtended)
   {
      /* only supply when supporting both ODS-2 and ODS-5 */
      sys$fao (&NumberFaoDsc, 0, &StringDsc, rqptr->PathOds);
      if (VMSnok (status =
          CgiVariable (rqptr, "PATH_ODS", String, VarType)))
         return (status);
   }

   if (VMSnok (status =
       CgiVariable (rqptr, "PATH_TRANSLATED",
                    rqptr->ParseOds.ExpFileName, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "QUERY_STRING", rqptr->rqHeader.QueryStringPtr, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "REMOTE_ADDR",
                    rqptr->rqNet.ClientIpAddressString, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "REMOTE_HOST", rqptr->rqNet.ClientHostName, VarType)))
      return (status);

   sys$fao (&NumberFaoDsc, 0, &StringDsc, rqptr->rqNet.ClientPort);
   if (VMSnok (status =
       CgiVariable (rqptr, "REMOTE_PORT", String, VarType)))
      return (status);

   if (rqptr->rqAuth.CaseLess)
   {
      /* force remote user name to upper-case if a case-less authentication */
      zptr = (sptr = String) + sizeof(String)-1;
      for (cptr = rqptr->RemoteUser;
           *cptr && sptr < zptr;
           *sptr++ = toupper(*cptr++));
      *sptr = '\0';
      if (VMSnok (status =
          CgiVariable (rqptr, "REMOTE_USER", String, VarType)))
         return (status);
   }
   else
   if (VMSnok (status =
       CgiVariable (rqptr, "REMOTE_USER", rqptr->RemoteUser, VarType)))
      return (status);

   if (rqptr->rqPathSet.CharsetPtr != NULL)
   {
      if (VMSnok (status =
          CgiVariable (rqptr, "REQUEST_CHARSET",
                       rqptr->rqPathSet.CharsetPtr, VarType)))
         return (status);
   }

   if (rqptr->rqPathSet.ContentTypePtr != NULL)
   {
      if (VMSnok (status =
          CgiVariable (rqptr, "REQUEST_CONTENT_TYPE",
                       rqptr->rqPathSet.ContentTypePtr, VarType)))
         return (status);
   }

   if (VMSnok (status =
       CgiVariable (rqptr, "REQUEST_METHOD", rqptr->rqHeader.MethodName, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "REQUEST_SCHEME",
                    rqptr->ServicePtr->RequestSchemeNamePtr, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "REQUEST_TIME_GMT", rqptr->rqTime.GmDateTime, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "REQUEST_TIME_LOCAL", LocalDateTime, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "REQUEST_URI", rqptr->rqHeader.RequestUriPtr, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "SCRIPT_NAME", rqptr->ScriptName, VarType)))
      return (status);

   /* Apache-like */
   if (rqptr->DclTaskPtr != NULL)
   {
      /* SSI "#dcl" will not always have a script file name */
      if (VMSnok (status =
           CgiVariable (rqptr, "SCRIPT_FILENAME",
                        rqptr->DclTaskPtr->ScriptFileName, VarType)))
          return (status);
   }

   /* Apache-like */
   if (Config.cfServer.AdminEmail[0])
      if (VMSnok (status =
          CgiVariable (rqptr, "SERVER_ADMIN",
                       Config.cfServer.AdminEmail, VarType)))
         return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "SERVER_CHARSET", Config.cfContent.CharsetDefault, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "SERVER_GMT", TimeGmtString, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "SERVER_NAME",
                    rqptr->ServicePtr->ServerHostName, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "SERVER_ADDR",
                    rqptr->ServicePtr->ServerIpAddressString, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "SERVER_PROTOCOL", HttpProtocol, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "SERVER_PORT",
                    rqptr->ServicePtr->ServerPortString, VarType)))
      return (status);

   /* Apache-like */
   ServerSignature (rqptr, String, sizeof(String));
   if (String[0])
      if (VMSnok (status =
          CgiVariable (rqptr, "SERVER_SIGNATURE", String, VarType)))
         return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "SERVER_SOFTWARE", SoftwareID, VarType)))
      return (status);

   /* Apache-like */
   if (VMSnok (status =
       CgiVariable (rqptr, "UNIQUE_ID", GenerateUniqueId(rqptr), VarType)))
      return (status);

   /*****************/
   /* SSL variables */
   /*****************/

   if (VMSnok (status = SesolaCgiGenerateVariables (rqptr, VarType)))
      return (status);

   /***************************/
   /* query string components */
   /***************************/

   if (rqptr->rqHeader.QueryStringPtr[0])
   {
      KeyCount = 0;

      cptr = rqptr->rqHeader.QueryStringPtr;
      while (*cptr && *cptr != '=') cptr++;
      /* if an equal symbol was found then its a form not a keyword search */
      if (*cptr)
      {
         /***************/
         /* form fields */
         /***************/

         memcpy (FormFieldName, "FORM_", 5);
         cptr = rqptr->rqHeader.QueryStringPtr;
         while (*cptr)
         {
            sptr = FormFieldName + 5;
            zptr = FormFieldName + sizeof(FormFieldName);
            while (*cptr && *cptr != '=' && *cptr != '&' && sptr < zptr)
            {
               if (isalnum(*cptr))
                  *sptr++ = toupper(*cptr++);
               else
               {
                  *sptr++ = '_';
                  cptr++;
               }
            }
            if (sptr >= zptr)
            {
               ErrorVmsStatus (rqptr, SS$_BUFFEROVF, FI_LI);
               return (STS$K_ERROR);
            }
            *sptr = '\0';

            if (!FormFieldName[5] || (*cptr && *cptr != '=' && *cptr != '&'))
            {
               if (*cptr == '&')
               {
                  /* allow back-to-back '&' (effectively null field name) */
                  cptr++;
                  continue;
               }
               /* error, out-of-place '=' (no field name) */
               rqptr->rqResponse.HttpStatus = 400;
               ErrorGeneral (rqptr, MsgFor(rqptr,MSG_REQUEST_URL_FORM), FI_LI);
               return (STS$K_ERROR);
            }

            /* if encountered an '=' */
            if (*cptr == '=') cptr++;
            Count = CgiUrlDecodeString (rqptr, cptr,
                                        String, sizeof(String), '&');
            if (Count < 0) return (STS$K_ERROR);
            cptr += Count;
            if (VMSnok (status =
                CgiVariable (rqptr, FormFieldName, String, VarType)))
               return (status);
         }

         if (VMSnok (status =
             CgiVariable (rqptr, "KEY_COUNT", "0", VarType)))
            return (status);
      }
      else
      {
         /******************/
         /* query keywords */
         /******************/

         cptr = rqptr->rqHeader.QueryStringPtr;
         while (*cptr)
         {
            sys$fao (&NumberFaoDsc, 0, &StringDsc, ++KeyCount);
            memcpy (FormFieldName, "KEY_", 4);
            strcpy (FormFieldName+4, String);

            Count = CgiUrlDecodeString (rqptr, cptr,
                                        String, sizeof(String), '+');
            if (Count < 0) return (STS$K_ERROR);
            cptr += Count;
            if (VMSnok (status =
                CgiVariable (rqptr, FormFieldName, String, VarType)))
               return (status);
         }

         sys$fao (&NumberFaoDsc, 0, &StringDsc, KeyCount);
         if (VMSnok (status =
             CgiVariable (rqptr, "KEY_COUNT", String, VarType)))
            return (status);
      }
   }
   else
   {
      /* no keywords if no query string! */
      if (VMSnok (status =
          CgiVariable (rqptr, "KEY_COUNT", "0", VarType)))
         return (status);
   }

   return (CgiVariable (rqptr, NULL, NULL, VarType));
}

/*****************************************************************************/
/*
URL-decode a URL-encoded string!  Return -1 if an error occurs, otherwise
return the number of characters read from the original, encoded string.
*/ 

int CgiUrlDecodeString
(
struct RequestStruct *rqptr,
char *EncodedString,
char *DecodedString,
int SizeofDecodedString,
char Separator
)
{
   register char  c;
   register char  *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CgiUrlDecodeString() |%s|\n", EncodedString);

   cptr = EncodedString;
   zptr = (sptr = DecodedString) + SizeofDecodedString;
   while (*cptr && *cptr != Separator && sptr < zptr)
   {
      if (*cptr == '+')
      {
         *sptr++ = ' ';
         cptr++;
      }
      else
      if (*cptr == '%')
      {
         /* an escaped character ("%xx" where xx is a hex number) */
         cptr++;
         c = 0;
         if (*cptr >= '0' && *cptr <= '9')
            { c = (*cptr - (int)'0') << 4; cptr++; }
         else
         if (tolower(*cptr) >= 'a' && tolower(*cptr) <= 'f')
            { c = (tolower(*cptr) - (int)'a' + 10) << 4; cptr++; }
         else
         {
            rqptr->rqResponse.HttpStatus = 400;
            ErrorGeneral (rqptr, MsgFor(rqptr,MSG_REQUEST_URL_ENC), FI_LI);
            return (-1);
         }
         if (*cptr >= '0' && *cptr <= '9')
            { c += (*cptr - (int)'0'); cptr++; }
         else
         if (tolower(*cptr) >= 'a' && tolower(*cptr) <= 'f')
            { c += (tolower(*cptr) - (int)'a' + 10); cptr++; }
         else
         {
            rqptr->rqResponse.HttpStatus = 400;
            ErrorGeneral (rqptr, MsgFor(rqptr,MSG_REQUEST_URL_ENC), FI_LI);
            return (-1);
         }
         if (sptr < zptr) *sptr++ = c;
      }
      else
         *sptr++ = *cptr++;
   }
   if (sptr >= zptr)
   {
      ErrorGeneral (rqptr, MsgFor(rqptr,MSG_GENERAL_OVERFLOW), FI_LI);
      return (-1);
   }
   *sptr = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", DecodedString);
   if (*cptr) cptr++;

   return (cptr - EncodedString);
}

/*****************************************************************************/
/*
Depending on whether the CGI variable is for creation at the DCL CLI or as part
of a CGI variable stream (e.g. CGIplus) create a null-terminated string
conatining either commands to create a DCL symbol, or a 'name=value' pair.

DCL symbol creation at the command line is limited by the CLI command line 
length (255 characters).  Symbol values however can be up to approximately 
1000 characters, probably enough for any CGI variable value.  If a CGI value 
is too large for for a single command-line assignment then build it up using 
multiple assignments, a symbol assignment kludge!
*/ 

int CgiVariable
(
struct RequestStruct *rqptr,
char *SymbolName,
char *SymbolValue,
int VarType
)
{
   static char  DeleteAllLocalSymbol[] = "DELETE/SYMBOL/LOCAL/ALL";

   register char  *cptr, *sptr, *zptr;

   int  status,
        SymbolCount,
        SymbolValueLength;
   unsigned short  Length;
   char  *CgiVariablePrefixPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "CgiVariable() |%s|%s|\n", SymbolName, SymbolValue);

   if (SymbolName == NULL)
   {
      /************************/
      /* end of CGI variables */
      /************************/

      if (rqptr->rqCgi.BufferRemaining < sizeof(short))
         CgiVariableBufferMemory (rqptr, sizeof(short));

      /* zero length record terminates generated CGI variables */
      *(short*)rqptr->rqCgi.BufferCurrentPtr = 0;
      rqptr->rqCgi.BufferRemaining -= sizeof(short);
      rqptr->rqCgi.BufferCurrentPtr += sizeof(short);

      if (Debug)
         fprintf (stdout, "buffer used: %d\n",
                  rqptr->rqCgi.BufferCurrentPtr - rqptr->rqCgi.BufferPtr);

      return (SS$_NORMAL);
   }

   if (rqptr->rqPathSet.CgiPrefixPtr == NULL)
      CgiVariablePrefixPtr = DclCgiVariablePrefix;
   else
      CgiVariablePrefixPtr = rqptr->rqPathSet.CgiPrefixPtr;

   if (SymbolValue == NULL) SymbolValue = "";

   if (rqptr->WatchItem &&
       (WatchEnabled & WATCH_CGI))
      WatchDataFormatted ("!AZ=!AZ\n", SymbolName, SymbolValue);

   if (VarType == CGI_VARIABLE_STREAM)
   {
      /******************************/
      /* variables in record stream */
      /******************************/

      if (rqptr->rqCgi.BufferRemaining <= DclCgiPlusInSize + sizeof(short))
         CgiVariableBufferMemory (rqptr, DclCgiPlusInSize*2);

      zptr = (sptr = rqptr->rqCgi.BufferCurrentPtr + sizeof(short)) +
             (rqptr->rqCgi.BufferRemaining - sizeof(short));
      for (cptr = CgiVariablePrefixPtr; *cptr; *sptr++ = *cptr++);
      for (cptr = SymbolName; *cptr && sptr < zptr; *sptr++ = *cptr++);
      if (sptr < zptr) *sptr++ = '=';
      for (cptr = SymbolValue; *cptr && sptr < zptr; *sptr++ = *cptr++);
      if (sptr < zptr) *sptr++ = '\0';
      if (sptr < zptr)
      {
         Length = sptr - (char*)rqptr->rqCgi.BufferCurrentPtr - sizeof(short);
         *(short*)rqptr->rqCgi.BufferCurrentPtr = Length;
         rqptr->rqCgi.BufferRemaining -= Length + sizeof(short);
         rqptr->rqCgi.BufferCurrentPtr += Length + sizeof(short);
         return (SS$_NORMAL);
      }
      ErrorGeneral (rqptr, MsgFor(rqptr,MSG_GENERAL_OVERFLOW), FI_LI);
      return (STS$K_ERROR);
   }

   /********************************/
   /* variables created at DCL CLI */
   /********************************/

   if (rqptr->rqCgi.BufferRemaining <= 256 + sizeof(short))
      CgiVariableBufferMemory (rqptr, DclSysCommandSize);

   zptr = (sptr = rqptr->rqCgi.BufferCurrentPtr + sizeof(short)) + 255;
   for (cptr = CgiVariablePrefixPtr; *cptr; *sptr++ = *cptr++);
   for (cptr = SymbolName; *cptr && sptr < zptr; *sptr++ = *cptr++);
   if (sptr < zptr) *sptr++ = '=';
   if (sptr < zptr) *sptr++ = '=';
   if (sptr < zptr) *sptr++ = '\"';
   cptr = SymbolValue;
   while (*cptr && sptr < zptr)
   {
      if (*cptr == '\"')
      {
         /* escape double quotes, by doubling up */
         if (sptr+2 < zptr)
         {
            cptr++;
            *sptr++ = '\"';
            *sptr++ = '\"';
         }
         else
            zptr = sptr;
         continue;
      }

      if (*cptr == '\'')
      {
         /* escape single quotes by concatenation */
         if (sptr+7 < zptr)
         {
            cptr++;
            memcpy (sptr, "\"+\"\'\"+\"", 7);
            sptr += 7;
         }
         else
            sptr = zptr;
         continue;
      }

      if (sptr < zptr) *sptr++ = *cptr++;
   }
   if (sptr < zptr) *sptr++ = '\"';
   if (sptr < zptr)
   {
      *sptr++ = '\0';
      Length = sptr - (char*)rqptr->rqCgi.BufferCurrentPtr - sizeof(short);
      *(short*)rqptr->rqCgi.BufferCurrentPtr = Length;
      rqptr->rqCgi.BufferRemaining -= Length + sizeof(short);
      rqptr->rqCgi.BufferCurrentPtr += Length + sizeof(short);
      return (SS$_NORMAL);
   }

   /* loop assigning maximum amount allowed by DCL until all assigned */
   SymbolCount = 0;
   cptr = SymbolValue;
   while (*cptr)
   {
      if (rqptr->rqCgi.BufferRemaining <= 256 + sizeof(short))
         CgiVariableBufferMemory (rqptr, DclSysCommandSize);

      if (Debug) fprintf (stdout, "|%s|\n", cptr);

      /* set this to DCL line length minus one for the terminating quote */
      zptr = (sptr = rqptr->rqCgi.BufferCurrentPtr + sizeof(short)) + 254;
      switch (++SymbolCount)
      {
         case 1 : memcpy (sptr, "A=\"", 3); break;
         case 2 : memcpy (sptr, "B=\"", 3); break;
         case 3 : memcpy (sptr, "C=\"", 3); break;
         case 4 : memcpy (sptr, "D=\"", 3); break;
         case 5 : memcpy (sptr, "E=\"", 3); break;
         case 6 : memcpy (sptr, "F=\"", 3); break;
         default : memcpy (sptr, "X=\"", 3); break;
      }
      sptr += 3;
      while (*cptr && sptr < zptr)
      {
         if (*cptr == '\"')
         {
            /* escape double quotes, by doubling up */
            if (sptr+2 < zptr)
            {
               cptr++;
               *sptr++ = '\"';
               *sptr++ = '\"';
            }
            else
               zptr = sptr;
            continue;
         }

         if (*cptr == '\'')
         {
            /* escape single quotes by concatenation */
            if (sptr+7 < zptr)
            {
               cptr++;
               memcpy (sptr, "\"+\"\'\"+\"", 7);
               sptr += 7;
            }
            else
               sptr = zptr;
            continue;
         }

         if (sptr < zptr) *sptr++ = *cptr++;
      }

      *sptr++ = '\"';
      *sptr++ = '\0';
      Length = sptr - (char*)rqptr->rqCgi.BufferCurrentPtr - sizeof(short);
      *(short*)rqptr->rqCgi.BufferCurrentPtr = Length;
      rqptr->rqCgi.BufferRemaining -= Length + sizeof(short);
      rqptr->rqCgi.BufferCurrentPtr += Length + sizeof(short);
   }
   SymbolValueLength = cptr - SymbolValue;

   /* assign the temporary symbol value(s) to the CGI symbol */
   if (rqptr->rqCgi.BufferRemaining <= 256 + sizeof(short))
      CgiVariableBufferMemory (rqptr, DclSysCommandSize);
   sptr = rqptr->rqCgi.BufferCurrentPtr + sizeof(short);
   for (cptr = CgiVariablePrefixPtr; *cptr; *sptr++ = *cptr++);
   for (cptr = SymbolName; *cptr; *sptr++ = *cptr++);
   switch (SymbolCount)
   {
      case 1 : memcpy (sptr, "==A", 4); sptr += 4; break;
      case 2 : memcpy (sptr, "==A+B", 6); sptr += 6; break;
      case 3 : memcpy (sptr, "==A+B+C", 8); sptr += 8; break;
      case 4 : memcpy (sptr, "==A+B+C+D", 10); sptr += 10; break;
      case 5 : memcpy (sptr, "==A+B+C+D+E", 12); sptr += 12; break;
      case 6 : memcpy (sptr, "==A+B+C+D+E+F", 14); sptr += 14; break;
      default : break;
   }
   Length = sptr - (char*)rqptr->rqCgi.BufferCurrentPtr - sizeof(short);

   /* 1005 seems to be a MAGIC TOTAL LINE LENGTH (for this approach anyway) */
   if (SymbolCount > 6 ||
       Length + SymbolValueLength > 1006)
   {
      /* cannot create a symbol this size */
      rqptr->rqResponse.ErrorTextPtr = SymbolName;
      rqptr->rqResponse.ErrorOtherTextPtr = "DCL symbol assignment too large!";
      ErrorVmsStatus (rqptr, CLI$_BUFOVF, FI_LI);
      return (STS$K_ERROR);
   }

   *(short*)rqptr->rqCgi.BufferCurrentPtr = Length;
   rqptr->rqCgi.BufferRemaining -= Length + sizeof(short);
   rqptr->rqCgi.BufferCurrentPtr += Length + sizeof(short);

   /* not really necessary, but let's be tidy */
   if (rqptr->rqCgi.BufferRemaining <=
       sizeof(DeleteAllLocalSymbol) + sizeof(short))
      CgiVariableBufferMemory (rqptr, DclSysCommandSize);
   memcpy (rqptr->rqCgi.BufferCurrentPtr + sizeof(short),
           DeleteAllLocalSymbol, sizeof(DeleteAllLocalSymbol));
   *(short*)rqptr->rqCgi.BufferCurrentPtr = sizeof(DeleteAllLocalSymbol);
   rqptr->rqCgi.BufferRemaining -= sizeof(DeleteAllLocalSymbol) + sizeof(short);
   rqptr->rqCgi.BufferCurrentPtr += sizeof(DeleteAllLocalSymbol) + sizeof(short);

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Allocate, or reallocate on subsequent calls, memory for storing the CGI
variables in.  Adjusts the CGI variable buffer pointers, lengths, etc,
appropriately.
*/ 

CgiVariableBufferMemory
(
struct RequestStruct *rqptr,
int BufferChunk
)
{
   int  CurrentLength,
        CurrentOffset;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "CgiVariableBufferMemory() %d %d\n",
               rqptr->rqCgi.BufferLength, BufferChunk);

   if (!rqptr->rqCgi.BufferLength)
   {
      /* initialize CGI variable buffer */
      rqptr->rqCgi.BufferLength = rqptr->rqCgi.BufferRemaining = BufferChunk;
      rqptr->rqCgi.BufferPtr = rqptr->rqCgi.BufferCurrentPtr =
         VmGetHeap (rqptr, BufferChunk);
      return;
   }

   CurrentLength = rqptr->rqCgi.BufferLength;
   CurrentOffset = rqptr->rqCgi.BufferCurrentPtr - rqptr->rqCgi.BufferPtr;

   rqptr->rqCgi.BufferPtr =
      VmReallocHeap (rqptr, rqptr->rqCgi.BufferPtr,
                     CurrentLength + BufferChunk, FI_LI);

   rqptr->rqCgi.BufferLength = CurrentLength + BufferChunk;
   rqptr->rqCgi.BufferRemaining += BufferChunk;
   rqptr->rqCgi.BufferCurrentPtr = rqptr->rqCgi.BufferPtr + CurrentOffset;
}

/*****************************************************************************/
/*
Process output from a CGI script (or even plain DCL processing).  For CGI
scripts check the first output from each for CGI-relevant HTTP header lines. 
For all output check that carriage-control is as required.
*/ 

int CgiOutput
(
struct RequestStruct *rqptr,
char *OutputPtr,
int OutputCount
)
{
   int  value,
        HeaderLineCount;
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CgiOutput() %d\n", OutputCount);

   if (OutputPtr == NULL)
   {
      /********************/
      /* reset CGI output */
      /********************/

      rqptr->rqCgi.BufferRecords =
         rqptr->rqCgi.CalloutInProgress =
         rqptr->rqCgi.ContentTypeText =
         rqptr->rqCgi.IsCliDcl =
         rqptr->rqCgi.ProcessingBody =
         rqptr->rqCgi.RecordMode = false;

      rqptr->rqCgi.CalloutOutputCount =
         rqptr->rqCgi.HeaderLineCount =
         rqptr->rqCgi.ScriptRetryCount =
         rqptr->rqCgi.RecordCount =
         rqptr->rqCgi.XVMSRecordMode = 0;

      return (0);
   }

   rqptr->rqCgi.RecordCount++;

   if (rqptr->rqCgi.EofLength)
   {
      if (*(long*)OutputPtr == *(long*)rqptr->rqCgi.EofPtr &&
          OutputCount >= rqptr->rqCgi.EofLength &&
          OutputCount <= rqptr->rqCgi.EofLength+2 &&
          !memcmp (OutputPtr, rqptr->rqCgi.EofPtr,
                              rqptr->rqCgi.EofLength))
      {
         /******************************/
         /* end of output from script! */
         /******************************/

         if (!rqptr->rqCgi.IsCliDcl &&
             rqptr->rqCgi.RecordCount == 1 &&
             Config.cfScript.CgiStrictOutput)
         {
            /* hmmmm, didn't receive any output */
            if (rqptr->WatchItem && (WatchEnabled & WATCH_CGI))
               WatchThis (rqptr, FI_LI, WATCH_CGI, ErrorCgiNotStrict);
            if (Debug) fprintf (stdout, "NOT SCRICT\n");
            return (CGI_OUTPUT_NOT_STRICT);
         }

         if (Debug) fprintf (stdout, "OUTPUT EOF!\n");
         return (CGI_OUTPUT_END);
      }
   }

   if (rqptr->rqCgi.CalloutInProgress)
   {
      /********************************/
      /* output escape is in progress */
      /********************************/

      if (*(long*)OutputPtr == *(long*)rqptr->rqCgi.EotPtr &&
          OutputCount >= rqptr->rqCgi.EotLength &&
          OutputCount <= rqptr->rqCgi.EotLength+2 &&
          !memcmp (OutputPtr, rqptr->rqCgi.EotPtr, rqptr->rqCgi.EotLength))
      {
         /******************************/
         /* end of escape from script! */
         /******************************/

         if (Debug) fprintf (stdout, "OUTPUT EOT!\n");
         return (CGI_OUTPUT_ESCAPE_END);
      }

      return (CGI_OUTPUT_ESCAPE);
   }

   if (rqptr->rqCgi.EscLength)
   {
      if (*(long*)OutputPtr == *(long*)rqptr->rqCgi.EscPtr &&
          OutputCount >= rqptr->rqCgi.EscLength &&
          OutputCount <= rqptr->rqCgi.EscLength+2 &&
          !memcmp (OutputPtr, rqptr->rqCgi.EscPtr, rqptr->rqCgi.EscLength))
      {
         /******************************/
         /* escape from script output! */
         /******************************/

         if (Debug) fprintf (stdout, "OUTPUT ESC!\n");
         return (CGI_OUTPUT_ESCAPE_BEGIN);
      }
   }

   if (rqptr->rqResponse.LocationPtr != NULL)
   {
      /************/
      /* redirect */
      /************/

      /* absorb any output from script until the request processing ends */
      return (CGI_OUTPUT_ABSORB);
   }

   HeaderLineCount = 0;
   if (!rqptr->rqCgi.ProcessingBody &&
       !rqptr->rqCgi.IsCliDcl)
   {
      /*************************/
      /* processing CGI header */
      /*************************/

      /* look at each line in the script-generated CGI response header */
      cptr = OutputPtr;
      while (*cptr || !OutputCount)
      { 
         HeaderLineCount++;
         value = CgiHeader (rqptr, cptr, OutputCount);

         if (value == CGI_OUTPUT_HTTP)
         {
            if (rqptr->WatchItem && (WatchEnabled & WATCH_CGI))
               WatchThis (rqptr, FI_LI, WATCH_CGI,
                          "RESPONSE stream mode (HTTP)");
            rqptr->rqCgi.ProcessingBody = true;
            return (OutputCount);
         }

         if (value == CGI_OUTPUT_END ||
             value == CGI_OUTPUT_NOT_STRICT ||
             value == CGI_OUTPUT_ABSORB)
            return (value);

         if (value == CGI_OUTPUT_END_HEADER ||
             value == CGI_OUTPUT_PLAIN_TEXT)
         {
            if (rqptr->rqCgi.XVMSRecordMode)
               rqptr->rqCgi.RecordMode = rqptr->rqCgi.XVMSRecordMode-1;
            else
            if (rqptr->rqCgi.ContentTypeText)
               rqptr->rqCgi.RecordMode = true;
            else
               rqptr->rqCgi.RecordMode = false;
            if (Debug)
               fprintf (stdout,
                  "ContentTypeText: %d XVMSRecordMode: %d RecordMode: %d\n",
                   rqptr->rqCgi.ContentTypeText,
                   rqptr->rqCgi.XVMSRecordMode-1,
                   rqptr->rqCgi.RecordMode);

            if (rqptr->WatchItem && (WatchEnabled & WATCH_CGI))
               WatchThis (rqptr, FI_LI, WATCH_CGI, "RESPONSE !AZ mode",
                          rqptr->rqCgi.RecordMode ? "record" : "stream");

            /* start processing the body now */
            rqptr->rqCgi.ProcessingBody = true;
            break;
         }

         /* scan to the start of the/any next line */
         while (*cptr && *cptr != '\n' && *(unsigned short*)cptr != '\r\n')
            cptr++;
         if (*cptr == '\r') cptr++;
         if (*cptr == '\n') cptr++;
      }
   }

   if (rqptr->rqCgi.RecordMode ||
       HeaderLineCount == 1)
   {
      /*************************/
      /* HTTP carriage-control */
      /*************************/

      if (Debug) fprintf (stdout, "rqptr->rqCgi.RecordMode\n");

      if (OutputPtr[OutputCount] != '\n')
      {
         OutputPtr[OutputCount++] = '\n';
         OutputPtr[OutputCount] = '\0';
      }
   }

   return (OutputCount);
}

/*****************************************************************************/
/*
For CGI-compliance the first line of the header should be a "Content-Type:", 
"Location:" or "Status:".  If a "Content-Type:" prepend a "200" header status 
line.  If a "Status:" use this to create a unique header status line.  If a 
"Location:" redirection header line call a function to establish the level of 
redirection, and do any associated processing.  For non-parse-header style CGI 
scripts check if the first line is a full HTTP status line.  If it is then do 
nothing, the script will supply the raw HTTP stream.  If none of the above 
then consider the script will supply nothing but plain text, none of the HTTP 
header, etc., and prepend a "200" status line and a "Content-Type: text/plain" 
line to create a complete HTTP header.  This is an extension to CGI behaviour, 
but makes simple, plain-text DCL scripts very easy!  Return true any action
required has been taken by this function, false if further processing is
required by the calling function.
*/ 

int CgiHeader 
(
struct RequestStruct *rqptr,
char *LinePtr,
int LineLength
)
{
   int  status;
   char  *cptr, *crptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CgiHeader()\n|%s|\n", LinePtr);

   rqptr->rqCgi.HeaderLineCount++;

   if (rqptr->WatchItem && (WatchEnabled & WATCH_CGI))
   {
      for (cptr = sptr = LinePtr;
           *cptr && *cptr != '\n' && *(unsigned short*)cptr != '\r\n';
           cptr++);
      switch (*cptr)
      {
         case '\0' : crptr= "\\0"; break;
         case '\n' : crptr= "\\n"; break;
         case '\r' : crptr= "\\r\\n";
      }
      WatchThis (rqptr, FI_LI, WATCH_CGI,
                 "RESPONSE header line !UL \'!AZ\' !UL bytes",
                 rqptr->rqCgi.HeaderLineCount, crptr, cptr-sptr);
      WatchDataFormatted ("!#AZ\n", cptr-sptr, sptr);
   }

   cptr = LinePtr;

   if (rqptr->rqCgi.HeaderLineCount == 1 &&
       *cptr == 'H' && !memcmp (cptr, "HTTP/", 5))
   {
      /*************************************************/
      /* script is supplying the full HTTP data stream */
      /*************************************************/

      if (Debug) fprintf (stdout, "(HTTP data stream)\n");

      /* a simple way to generate 0.9 responses, just for testing purposes! */
      if (!memcmp (cptr+4, "/0.9", 4)) *(long*)cptr = 'PTTH';

      /* get the response status code for logging purposes */
      while (*cptr && !ISLWS(*cptr)) cptr++;
      while (*cptr && !isdigit(*cptr)) cptr++;
      if (isdigit(*cptr)) rqptr->rqResponse.HttpStatus = atoi(cptr);
      return (CGI_OUTPUT_HTTP);
   }

   if (rqptr->rqCgi.HeaderLineCount > 1 &&
       (*cptr == '\n' || *(unsigned short*)cptr == '\r\n' || !LineLength))
   {
      /***************************/
      /* end of response header! */
      /***************************/

      return (CGI_OUTPUT_END_HEADER);
   }

   if (toupper(*cptr) == 'C' && strsame (cptr, "Content-Type:", 13))
   {
      /**************************/
      /* content-type specified */
      /**************************/

      /*
         For CGI content types of "text/..." add carriage control to each
         record needing it.  For non-text types this carriage-control is
         added to response header lines then turned off for the body.
      */

      cptr += 13;
      while (ISLWS(*cptr) && NOTEOL(*cptr)) cptr++;
      if (strsame (cptr, "text/", 5))
      {
         /* make sure each record received has HTTP carriage control */
         if (Debug) fprintf (stdout, "(RMS record stream)\n");
         rqptr->rqCgi.RecordMode = rqptr->rqCgi.ContentTypeText = true;
      }
      else
         rqptr->rqCgi.RecordMode = false;

      if (!rqptr->rqResponse.HeaderLength)
      {
         HttpHeader (rqptr, 0, NULL, -1, NULL, NULL);
         /* remove generated empty header line, script supplies rest of lines */
         if (rqptr->rqResponse.HeaderLength > 2)
         {
            rqptr->rqResponse.HeaderPtr[rqptr->rqResponse.HeaderLength-2] = '\0';
            rqptr->rqResponse.HeaderLength -= 2;
         }
      }

      return (LineLength);
   }

   if (toupper(*cptr) == 'S' && strsame (cptr, "Status:", 7))
   {
      /********************/
      /* status specified */
      /********************/

      /* make sure each record received has HTTP carriage control */
      if (Debug) fprintf (stdout, "(RMS record stream)\n");
      rqptr->rqCgi.RecordMode = true;

      /* create HTTP header status line using supplied status */
      cptr += 7;
      while (*cptr++ && !isdigit(*cptr) && *cptr != '\n') cptr++;
      /* get the response status code for header and logging purposes */
      if (isdigit(*cptr)) rqptr->rqResponse.HttpStatus = atoi(cptr);

      if (!rqptr->rqResponse.HeaderLength)
      {
         HttpHeader (rqptr, 0, NULL, -1, NULL, NULL);
         /* remove generated empty header line, script supplies rest of lines */
         if (rqptr->rqResponse.HeaderLength > 2)
         {
            rqptr->rqResponse.HeaderPtr[rqptr->rqResponse.HeaderLength-2] = '\0';
            rqptr->rqResponse.HeaderLength -= 2;
         }
      }

      return (LineLength);
   }

   if (toupper(*cptr) == 'L' && strsame (cptr, "Location:", 9))
   {
      /*******************/
      /* redirection URL */
      /*******************/

      cptr += 9;
      /* locate the first character of the "Location:" URL */
      while (ISLWS(*cptr)) cptr++;
      sptr = cptr;
      while (*cptr && !ISLWS(*cptr) && *cptr != '\r' && *cptr != '\n') cptr++;
      rqptr->rqResponse.LocationPtr = VmGetHeap (rqptr, cptr-sptr+1);
      memcpy (rqptr->rqResponse.LocationPtr, sptr, cptr-sptr);
      rqptr->rqResponse.LocationPtr[cptr-sptr] = '\0';
      if (Debug) fprintf (stdout, "LocationPtr |%s|\n", rqptr->rqResponse.LocationPtr);
      return (CGI_OUTPUT_ABSORB);
   }

   if (toupper(*cptr) == 'X' && strsame (cptr, "X-vms-record-mode:", 18))
   {
      /*****************************/
      /* VMS Apache record control */
      /*****************************/

      /*
         VMS Apache uses an eXtension header field so that the script
         can convey to the server whether the output should be processed
         as a binary stream or as records (the default).
      */

      cptr += 18;
      while (ISLWS(*cptr) && NOTEOL(*cptr)) cptr++;
      /* this value should either 0 or 1 ("no" or "yes") plus 1 */
      rqptr->rqCgi.XVMSRecordMode = atoi(cptr)+1;

      return (LineLength);
   }

   /****************/
   /* other output */
   /****************/

   /* only if it's not the first line do we continue on! */
   if (rqptr->rqCgi.HeaderLineCount > 1) return (LineLength);

   if (Config.cfScript.CgiStrictOutput)
   {
      if (rqptr->WatchItem && (WatchEnabled & WATCH_CGI))
         WatchThis (rqptr, FI_LI, WATCH_CGI, ErrorCgiNotStrict);
      return (CGI_OUTPUT_NOT_STRICT);
   }
   /* else drop through for non-strict (plain-text) output */

   /******************/
   /* non-CGI output */
   /******************/

   /* make sure each record received has HTTP carriage control */
   if (Debug) fprintf (stdout, "(RMS record stream)\n");
   rqptr->rqCgi.ContentTypeText = true;

   HttpHeader (rqptr, 0, "text/plain", -1, NULL, NULL);

   return (CGI_OUTPUT_PLAIN_TEXT);
}

/*****************************************************************************/
/*
Generate a (hopefully) unique sequence of 224 bits (28 bytes, which is a fair
bit :^) The string is generated using the a count sequence and quadword binary
times each time it is called.  This provides a continuous, non-repeating series
of unlikely bit combinations with a one in 2^224 chance (I think!) of presence
in an I/O stream.
*/ 

CgiSequence
(
char OneChar,
char *CgiSequencePtr,
int *CgiSequenceLengthPtr
)
{                                           
   static $DESCRIPTOR (SequenceFaoDsc, "^!AZ-!8XL!8XL!8XL-\0");
   static $DESCRIPTOR (SequenceDsc, "");
   static char  OneCharString [] = " ";
   static int  SequenceCount = 0;

   unsigned short  Length;
   unsigned long  BinTime[2];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CgiSequence()\n");

   SequenceCount++;
   sys$gettim(&BinTime);
   SequenceDsc.dsc$w_length = 29;
   SequenceDsc.dsc$a_pointer = CgiSequencePtr;
   OneCharString[0] = OneChar;
   sys$fao (&SequenceFaoDsc, &Length, &SequenceDsc,
            OneCharString, SequenceCount, BinTime[1], BinTime[0]);
   *CgiSequenceLengthPtr = Length - 1;

   if (Debug)
      fprintf (stdout, "%d |%s|\n", *CgiSequenceLengthPtr, CgiSequencePtr);
}

/*****************************************************************************/

